#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#######################################################################
##                                                                   ##
##                              pdb4all                              ##
##                                                                   ##
#######################################################################
#
#
# Copyright (C) 2021, Sergio Boneta
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see https://www.gnu.org/licenses/
#
#
# Check for updates:  https://github.com/boneta/pdb4all


"""
  =======
  pdb4all
  =======

  Protein conversion between common pdb formats and name conventions


"""

__version__ = '0.4.3'

from pdb4all.pdb import *
